import configparser
import pathlib
import re
import time
import warnings
from distutils import log

import setuptools


def _get_package_name_from_setup_config() -> str:
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    return config["metadata"]["name"]


def _find_version() -> str:
    """Get the version string from the top-level __init__.py file."""
    package_name = _get_package_name_from_setup_config()
    top_level_init_file_path = pathlib.Path(package_name, "__init__.py")
    with top_level_init_file_path.open(encoding="utf8") as f:
        version_match = re.search(r"__version__ = [\"\'](.*?)[\"\']", f.read())
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string in", top_level_init_file_path)


def _get_git_version_classifier() -> str:
    """
    Return the 'Git version' string to add to the setuptools 'classifiers' section.

    This is a non-standard classifier, used to track the git-hash in optellum's packages
    uploaded to the nexus repo.
    The classifiers are added to the PKG-INFO file created during the wheel-build.
    """
    root = str(pathlib.Path(__file__).parent)
    try:
        import git

        repo = git.Repo(path=root, search_parent_directories=True)
        commit = repo.head.object
        sha = commit.hexsha
        dt = time.strftime("%a, %d %b %Y %H:%M", time.gmtime(commit.committed_date))
        dirt = {False: "CLEAN", True: "DIRTY"}[repo.is_dirty()]
        git_version = f"SHA= {sha} Status= {dirt} (Date {dt})"
        if repo.is_dirty():
            warnings.warn(
                "WARNING: Dirty repo. Changed files:"
                + " ".join(item.a_path for item in repo.index.diff(None))
            )
    except ModuleNotFoundError:
        git_version = "Missing gitpython package"
        warnings.warn("Missing git package. Try:\n  pip install gitpython")

    except git.exc.InvalidGitRepositoryError as ex:
        git_version = f"Missing git repo at '{root}'"
        warnings.warn("InvalidGitRepositoryError: " + str(ex))

    git_string = f"Git Version :: {git_version}"
    log.info(git_string)

    return git_string


setuptools.setup(
    version=_find_version(),
)
